import subprocess
import unittest

from mock import patch

import tenper


class TenperTests(unittest.TestCase):
    def test_command_list(self):
        self.assertEqual(
            tenper.command_list('ls'),
            ['ls'])

        self.assertEqual(
            tenper.command_list('ls /'),
            ['ls', '/'])

        self.assertEqual(
            tenper.command_list('echo {message}', message='Hello, world.'),
            ['echo', 'Hello, world.'])


    def test_run(self):
        with patch('subprocess.call') as call:
            tenper.run('ls /')
            call.assert_called_once_with(['ls', '/'])

        with patch('subprocess.call') as call:
            tenper.run(
                'tmux new-window -t {session} -n {window}',
                session='some thing',
                window='hi')
            call.assert_called_once_with(
                ['tmux', 'new-window', '-t', 'some thing', '-n', 'hi'])


    def test_parse_args(self):
        f, a = tenper.parse_args(['edit', 'foo'])
        self.assertEqual(f, tenper.edit)
        self.assertEqual(a, 'foo')

        f, a = tenper.parse_args(['rebuild', 'bar'])
        self.assertEqual(f, tenper.rebuild)
        self.assertEqual(a, 'bar')

        f, a = tenper.parse_args(['delete', 'foo'])
        self.assertEqual(f, tenper.delete)
        self.assertEqual(a, 'foo')

        f, a = tenper.parse_args(['list'])
        self.assertEqual(f, tenper.list_envs)
        self.assertEqual(a, 'list')

        f, a = tenper.parse_args(['foobar'])
        self.assertEqual(f, tenper.start)
        self.assertEqual(a, 'foobar')



#!/usr/bin/env python
"""Tenper is a tmux wrapper with support for Python virtualenv.

The name is a corruption of gibberish:
    tmuxvirtualenvwrapper -> tenvpapper -> tenper.

Environment variables:
    TENPER_VERBOSE: Set to 'true' if you want to see the commands we execute.
    TENPER_CONFIGS: The path to where you want the configuration files stored.
        Defaults to ~/.tenper.
    TENPER_VIRTUALENVS: The path to where you keep your virtualenvs. Defaults
        to virtualenvwrapper's default of ~/.virtualenvs.
    TENPER_TMUX_COMMAND: Defaults to 'tmux'. Try 'tmux -2' if you want 256
        colors without TERM wrangling.
"""

import argparse
import contextlib
import os
import re
import shutil
import subprocess
import sys

import yaml

from . import command
from . import config


_print_commands = os.getenv('TENPER_VERBOSE', 'false') == 'true'

_run_context = {
    'editor': os.getenv('EDITOR'),
    'config_path': os.getenv('TENPER_CONFIGS') or \
        os.path.join(os.path.expanduser('~'), '.tenper'),
    'virtualenvs_path': os.getenv('TENPER_VIRTUALENVS') or \
        os.path.join(os.path.expanduser('~'), '.virtualenvs'),
    'tmux_command': os.getenv('TENPER_TMUX_COMMAND', 'tmux'),

    # These are the YAML config properties we'll want to use in the subprocess
    # calls.
    'config_file_name': None,
    'project_root': None,
    'session_name': None,
    'virtualenv_configured': False,
    'virtualenv_path': None,
    'virtualenv_python_binary': None,
    'virtualenv_use_site_packages': '--no-site-packages',

    # TODO(mason): Am I confounding two things here? The following are stored
    # from the config and used as part of the program logic, not simple
    # replacements applied to the subprocess calls. This is convenient, but I'm
    # unsure.
    'environment': {},
    'windows': [],
}


@contextlib.contextmanager
def run_context(**kwargs):
    """Updates the global run context safely."""

    global _run_context

    old_run_context = _run_context
    _run_context = old_run_context.copy()
    _run_context.update(**kwargs)
    yield _run_context
    _run_context = old_run_context


def configured_string(string):
    """Returns 'string' formatted with the run context."""
    return string.format(**_run_context)


def configured(key, default=None):
    """Returns a configuration parameter."""
    # This function is a little silly. I'm eliminating the references to
    # core._run_context in the other modules.
    return _run_context.get(key, default)


def run(command, interactive=False, **kwargs):
    """Runs a command. The command is formatted with the run_context.

    This permits the following usage. The run context is augmented with
    temporary parameters. We no longer need to execute string.format for every
    parameterized command. It's more legible.

        with run_context(temporary_thing='foobar'):
            run('cp {temporary_thing} {config_path}')

    Args:
        command: A string with spaces. It'll be split on spaces and then be
            formatted with the run_context and kwargs.
        kwargs: Additional formatting.

    Returns:
        A boolean indicating success and the command output: (ok, output).
    """

    formatting = dict(list(_run_context.items()) + list(kwargs.items()))
    command_list = []

    # The specific check for the tmux executable allows the
    # TENPER_TMUX_COMMAND to contain arguments. I always use tmux -2 to
    # force 256 colors because I don't like the TERM wrangling usually
    # advised.
    for part in command.split(' '):
        if part == '{tmux_command}':
            command_list.extend(_run_context['tmux_command'].split(' '))
        else:
            command_list.append(part.format(**formatting))

    if _print_commands:
        print('* {}'.format(' '.join(command_list)))

    # If you're inside a tmux session, tmux won't let you send commands to
    # another tmux session (even if it's detached). To force it, tmux suggests
    # unsetting the TMUX environment variable. We'll run our commands in a
    # modified environment with no TMUX set.
    env_no_tmux = os.environ.copy()
    old_tmux = env_no_tmux.pop('TMUX', None)

    # This is for the 'edit' command. I think it's the only way to open an
    # editor.
    if interactive:
        response = subprocess.call(command_list, env=env_no_tmux)
        ok = response == 0
        output = None

    else:
        try:
            output = subprocess.check_output(command_list, env=env_no_tmux)
            ok = True
        except subprocess.CalledProcessError as e:
            output = e.output
            ok = False

        output = output.decode('utf-8')

    return (ok, output)


def parse_args(args):
    """Handles user input.

    Args:
        args: Probably never anything but sys.argv[1:].

    Returns:
        A Namespace with 'command' and/or 'project_name' properties.
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=(
            'A wrapper for tmux sessions and (optionally) virtualenv.\n\n'
            'Usage:\n'
            '    tenper list\n'
            '    tenper edit my-project\n'
            '    tenper rebuild my-project\n'
            '    tenper delete my-project\n'
            '    tenper my-project\n'))

    if len(args) == 1:
        # 'list', 'completions', or a project name.
        parser.add_argument('project_name')
        parser.set_defaults(command='start')

    else:
        # Subcommand.
        subparsers = parser.add_subparsers(dest='command')

        def mksubparser(name, help_text):
            sp = subparsers.add_parser(name, help=help_text)
            sp.add_argument('project_name')

        mksubparser('edit', 'Edit a project\'s configuration.')
        mksubparser('rebuild', 'Delete an existing virtualenv and start a new one.')
        mksubparser('delete', 'Delete a project\'s virtualenv and configuration.')

    parsed_args = parser.parse_args(args)

    # meh.
    if parsed_args.project_name in ['list', 'completions']:
        parsed_args.command = parsed_args.project_name
        del parsed_args.project_name


    return parsed_args


def user_input(prompt):
    """Returns user input in Python 2 or 3."""

    try:
        return raw_input(prompt)
    except (EOFError, NameError, ValueError):
        return input(prompt)


def main(*args, **kwargs):
    arguments = parse_args(sys.argv[1:])

    if arguments.command == 'list':
        command.list()

    elif arguments.command == 'completions':
        command.completions()

    else:
        config_file_name = os.path.join(configured('config_path'),
                                        '{}.yml'.format(arguments.project_name))

        with run_context(**config.load(config_file_name)):
            getattr(command, arguments.command)(arguments.project_name)



import os
import shutil
import sys

from . import config


def _attach_or_switch():
    from . import core

    if os.getenv('TMUX'):
        core.run('{tmux_command} switch-client -t {session_name}')
    else:
        core.run('{tmux_command} attach-session -t {session_name}')


def _confirm_virtualenv(env):
    """Makes sure we have a virtualenv installed for env."""
    from . import core

    directory = core.configured('virtualenv_path')

    if (core.configured('virtualenv_configured') and
        directory and
        not os.path.exists(directory)):

        print(
            'We have a virtualenv confirgured for {}. Building...'.format(env))

        core.run((
            'virtualenv -p {virtualenv_python_binary} '
            '{virtualenv_use_site_packages} {virtualenv_path}'))


def _remove_virtualenv(env):
    """Deletes a possibly extant virtualenv and rebuild it."""
    from . import core

    # Short circuit; no virtualenv configured.
    if not core.configured('virtualenv_configured'):
        print('No virtualenv configured for {}.'.format(env))
        return

    directory = core.configured('virtualenv_path')

    if not os.path.exists(directory):
        print('No virtualenv created for {}.'.format(env))
        return

    response = core.user_input(
        'Are you sure you want to delete {}? '.format(directory))

    if response.strip() in ['yes', 'YES', 'y', 'Y']:
        shutil.rmtree(directory)
        print('Deleted {}.'.format(directory))
    else:
        print('Skipping.')


def _tmux_option(name, cast=str, default=None):
    """Returns the value of a tmux option or the value of default."""

    from . import core
    _, output = core.run('{tmux_command} show-options -g -t {session_name}')
    return _tmux_option_parser(name, output, cast, default)


def _tmux_option_parser(name, output, cast, default):
    """Returns the value from tmux's option output or the value of default."""

    for line in output.splitlines():
        parts = line.split()
        if parts[0] == name:
            return cast(' '.join(parts[1:]))

    return default


def _tmux_window_option(name, cast=str, default=None):
    """Returns the value of a tmux window option or the value of default."""

    from . import core
    _, output = core.run('{tmux_command} show-window-options -g -t {session_name}')
    return _tmux_option_parser(name, output, cast, default)


def completions():
    """Returns a space-separated list of available commands an arguments. This
    is appropriate for the zsh completions."""

    from . import core

    args = ['list', 'edit', 'rebuild', 'delete']

    if os.path.exists(core.configured('config_path')):
        for f in os.listdir(core.configured('config_path')):
            if f.endswith('.yml'):
                args.append(f[0:-4])

    print(' '.join(args))


def delete(env):
    """Removes an environment configuration and extant virtualenv."""
    from . import core

    _remove_virtualenv(env)

    file_name = core.configured('config_file_name')
    directory = os.path.dirname(file_name)

    if os.path.exists(file_name):
        print('Removing environment config at {}'.format(file_name))
        os.remove(file_name)
    else:
        print('No tenper config for {}.'.format(env))

    # Clean up the .tenper directory is it's no longer used.
    try:
        os.rmdir(os.path.dirname(core.configured('config_file_name')))
        print('Cleaned up tenper config directory at {}.'.format(directory))
    except OSError:
        pass


def edit(env):
    """Edit (or create) an environment's configuration."""
    from . import core

    config_file_name = core.configured('config_file_name')

    if not os.path.exists(config_file_name):
        config.create(config_file_name, env)

    with core.run_context():
        core.run('{editor} {config_file_name}', interactive=True)


def list():
    """Prints the configuration files in {config_path} to stdout."""
    from . import core

    directory = core.configured('config_path')

    # Short circuit; no configuration.
    if not os.path.exists(directory):
        print(core.configured_string(
            'You have no environments; {config_path} is empty.'))
        return

    args = [f[0:-4] for f in os.listdir(directory) if f.endswith('.yml')]

    if not args:
        print('None.')
    else:
        for yml in args:
            print(yml)


def rebuild(env):
    """Deletes a possibly extant virtualenv and rebuilds it."""

    _remove_virtualenv(env)
    _confirm_virtualenv(env)


def start(env):
    from . import core
    print('Starting {}'.format(env))
    _confirm_virtualenv(env)

    # Short circuit; prexisting session.
    ok, _ = core.run('{tmux_command} has-session -t {session_name}')
    if ok:
        core.user_input('This session already exists. Press any key to reattach.')
        _attach_or_switch()
        return

    core.run('{tmux_command} new-session -d -s {session_name}')
    core.run('{tmux_command} set-option -t {session_name} default-path {project_root}')

    status_left_length = _tmux_option('status-left-length', cast=int, default=0)
    if len(core.configured('session_name')) > status_left_length:
        core.run('{tmux_command} set-option -t {session_name} status-left-length ' +
                    str(len(core.configured('session_name'))))

    if core.configured('virtualenv_configured'):
        core.run(('{tmux_command} set-environment -t {session_name} '
                  'TENPER_VIRTUALENV {virtualenv_path}/bin/activate'))

    if core.configured('environment'):
        for k, v in core.configured('environment').items():
            with core.run_context(key=k, value=os.path.expandvars(v)):
                core.run(('{tmux_command} set-environment -t {session_name} '
                          '{key} {value}'))

    base_window_index = _tmux_option('base-index', cast=int, default=0)
    base_pane_index = _tmux_window_option('pane-base-index',
                                          cast=int,
                                          default=0)

    for window_index, window in enumerate(core.configured('windows', [])):
        with core.run_context(window_index=base_window_index+window_index):
            core.run(('{tmux_command} new-window -d -k -t '
                      '{session_name}:{window_index} -n {window_name}'),
                     window_name=window.get('name', 'No Name'))

            for pane_index, pane in enumerate(window.get('panes', [])):
                with core.run_context(pane_index=base_pane_index+pane_index,
                                      previous_pane_index=base_pane_index+pane_index-1):
                    if pane_index != 0:
                        core.run(('{tmux_command} split-window -t '
                                  '{session_name}:{window_index}.{previous_pane_index}'))

                    if core.configured('virtualenv_configured'):
                        core.run(('{tmux_command} send-keys -t '
                                  '{session_name}:{window_index}.{pane_index} '
                                  '{source_virtualenv} ENTER'),
                                 source_virtualenv='source $TENPER_VIRTUALENV')

                    # It might be an empty command.
                    if pane:
                        core.run(('{tmux_command} send-keys -t '
                                  '{session_name}:{window_index}.{pane_index} '
                                  '{pane_command} ENTER'),
                                 pane_command=pane)

            if window.get('layout'):
                core.run(('{tmux_command} select-layout -t '
                          '{session_name}:{window_index} {layout}'),
                         layout=window['layout'])

    core.run(('{tmux_command} select-window -t '
              '{session_name}:{base_window_index}'),
             base_window_index=base_window_index)

    core.run(('{tmux_command} select-pane -t '
              '{session_name}:{base_window_index}.{base_pane_index}'),
             base_window_index=base_window_index,
             base_pane_index=base_pane_index)

    _attach_or_switch()



from setuptools import setup, find_packages

setup(
    name='tenper',
    version='1.2.1',
    description='A tmux session manager with optional virtualenv support.',
    long_description=(
        'Tenper is a tmux wrapper. It provides project-based tmux window/pane '
        'layouts.  It has optional support for Python\'s virtualenv and the '
        'conventions it uses permits concurrent usage of virtualenvwrapper.'),
    keywords='tmux, virtualenv, virtualenvwrapper',
    author='Mason Staugler',
    author_email='mason@staugler.net',
    url='https://github.com/mqsoh/tenper',
    license='MIT license',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
    ],

    install_requires=['pyyaml'],
    packages=find_packages(exclude=['*_tests.py']),
    entry_points={'console_scripts': ['tenper = tenper.core:main']},
    scripts=['scripts/tenper-completion.sh'],
    package_data={'tenper': ['config_template.yml']})



import os

import yaml


def _get(config, key, default=None):
    """Gets and expands a configuration value.

    Returns:
        The value or None.
    """

    value = config.get(key, default)

    try:
        return os.path.expandvars(value)
    except:
        return value


def _get_virtualenv(config, key, default=None):
    """Gets and expands a value inside the 'virtualenv' in 'config'.

    The 'virtualenv' might not be available, so we want to abstract that test.

    Returns:
        The value or None.
    """

    value = config.get('virtualenv', {}).get(key, default)

    try:
        return os.path.expandvars(value)
    except:
        return value


def create(file_name, env):
    """Creates a new config file from the template.

    Args:
        file_name: The target config file name.
        env: The initial session name.

    Returns:
        A boolean indicating success.
    """

    directory = os.path.dirname(file_name)
    template_file_name = os.path.join(os.path.dirname(__file__),
                                      'config_template.yml')

    if not os.path.exists(directory):
        os.mkdir(directory)

    if not os.path.exists(file_name):
        # We could do a simple file copy if it weren't for prepopulating the
        # 'session name' in the config.
        with open(template_file_name, 'r') as template_file:
            with open(file_name, 'w') as config_file:
                config_file.write(template_file.read().format(env=env))

    return True


def load(file_name):
    """Returns a dictionary with a parsed config file.

    This will flatten the virtualenv definition for the core._run_context and
    apply os.path.expandvars where appropriate.

    Args:
        file_name: The path to a configuration file.

    Returns:
        If no file exists:
            {'config_file_name': file_name}

        If the file exists:
            {'config_file_name': file_name,
             'session_name': ...,
             'project_root': ...,
             'environment': ...,
             'windows': ...}

        If a virtualenv is configured the dict will also have:
            {'virtualenv_configured': True,
             'virtualenv_base_path': ...,
             'virtualenv_python_binary': ...,
             'virtualenv_use_site_packages': True | False}
    """

    from . import core

    # Short circuit; new environment.
    if not os.path.exists(file_name):
        return {'config_file_name': file_name}

    with open(file_name, 'r') as f:
        config = yaml.load(f)

    d = {'config_file_name': file_name,
         'session_name': _get(config, 'session name'),
         'project_root': _get(config, 'project root'),
         'environment': config.get('environment', None),
         'windows': config.get('windows', None)}

    if config.get('virtualenv'):
        site_packages = '--no-site-packages'
        if _get_virtualenv(config, 'site packages?'):
            site_packages = '--system-site-packages'

        d.update({'virtualenv_configured': True,
                  'virtualenv_path': os.path.join(core.configured('virtualenvs_path'), d['session_name']),
                  'virtualenv_python_binary': _get_virtualenv(config, 'python binary'),
                  'virtualenv_use_site_packages': site_packages})

    return d
